@@ -68,6 +68,12 @@ static TrieState * trie_state_new (const Trie *trie,
                                    short       suffix_idx,
                                    short       is_suffix);
 
+static Bool
+trie_store_conditionally (Trie            *trie,
+                          const AlphaChar *key,
+                          TrieData         data,
+                          Bool             is_overwrite);
+
 static Bool        trie_branch_in_branch (Trie           *trie,
                                           TrieIndex       sep_node,
                                           const TrieChar *suffix,
@@ -150,18 +156,42 @@ trie_new_from_file (const char *path)
     if (!trie_file)
         return NULL;
 
+    trie = trie_fread (trie_file);
+    fclose (trie_file);
+    return trie;
+}
+
+/**
+ * @brief Create a new trie by reading from an open file
+ *
+ * @param file  : the handle of the open file
+ *
+ * @return a pointer to the created trie, NULL on failure
+ *
+ * Create a new trie and initialize its contents by reading from the open
+ * @a file. After reading, the file pointer is left at the end of the trie data.
+ * This can be useful for reading embedded trie index as part of a file data.
+ *
+ * The created object must be freed with trie_free().
+ *
+ * Available since: 0.2.4
+ */
+Trie *
+trie_fread (FILE *file)
+{
+    Trie       *trie;
+
     trie = (Trie *) malloc (sizeof (Trie));
     if (!trie)
-        goto exit_file_openned;
+        return NULL;
 
-    if (NULL == (trie->alpha_map = alpha_map_read_bin (trie_file)))
+    if (NULL == (trie->alpha_map = alpha_map_fread_bin (file)))
         goto exit_trie_created;
-    if (NULL == (trie->da   = da_read (trie_file)))
+    if (NULL == (trie->da   = da_fread (file)))
         goto exit_alpha_map_created;
-    if (NULL == (trie->tail = tail_read (trie_file)))
+    if (NULL == (trie->tail = tail_fread (file)))
         goto exit_da_created;
 
-    fclose (trie_file);
     trie->is_dirty = FALSE;
     return trie;
 
@@ -171,8 +201,6 @@ exit_alpha_map_created:
     alpha_map_free (trie->alpha_map);
 exit_trie_created:
     free (trie);
-exit_file_openned:
-    fclose (trie_file);
     return NULL;
 }
 
@@ -214,29 +242,43 @@ trie_save (Trie *trie, const char *path)
     if (!file)
         return -1;
 
-    if (alpha_map_write_bin (trie->alpha_map, file) != 0) {
-        res = -1;
-        goto exit_file_openned;
-    }
-
-    if (da_write (trie->da, file) != 0) {
-        res = -1;
-        goto exit_file_openned;
-    }
-
-    if (tail_write (trie->tail, file) != 0) {
-        res = -1;
-        goto exit_file_openned;
-    }
-
-    trie->is_dirty = FALSE;
-    return 0;
-
-exit_file_openned:
+    res = trie_fwrite (trie, file);
     fclose (file);
     return res;
 }
 
+/**
+ * @brief Write trie data to an open file
+ *
+ * @param trie  : the trie
+ *
+ * @param file  : the open file
+ *
+ * @return 0 on success, non-zero on failure
+ *
+ * Write @a trie data to @a file which is opened for writing.
+ * After writing, the file pointer is left at the end of the trie data.
+ * This can be useful for embedding trie index as part of a file data.
+ *
+ * Available since: 0.2.4
+ */
+int
+trie_fwrite (Trie *trie, FILE *file)
+{
+    if (alpha_map_fwrite_bin (trie->alpha_map, file) != 0)
+        return -1;
+
+    if (da_fwrite (trie->da, file) != 0)
+        return -1;
+
+    if (tail_fwrite (trie->tail, file) != 0)
+        return -1;
+
+    trie->is_dirty = FALSE;
+
+    return 0;
+}
+
 /**
  * @brief Check pending changes
  *
@@ -310,7 +352,7 @@ trie_retrieve (const Trie *trie, const AlphaChar *key, TrieData *o_data)
 }
 
 /**
- * @brief Store a value for  an entry to trie
+ * @brief Store a value for an entry to trie
  *
  * @param trie  : the trie
  * @param key   : the key for the entry to retrieve
@@ -324,6 +366,39 @@ trie_retrieve (const Trie *trie, const AlphaChar *key, TrieData *o_data)
  */
 Bool
 trie_store (Trie *trie, const AlphaChar *key, TrieData data)
+{
+    return trie_store_conditionally (trie, key, data, TRUE);
+}
+
+/**
+ * @brief Store a value for an entry to trie only if the key is not present
+ *
+ * @param trie  : the trie
+ * @param key   : the key for the entry to retrieve
+ * @param data  : the data associated to the entry
+ *
+ * @return boolean value indicating the success of the process
+ *
+ * Store a @a data for the given @a key in @a trie. If @a key does not 
+ * exist in @a trie, it will be appended. If it does, the function will
+ * return failure and the existing value will not be touched.
+ *
+ * This can be useful for multi-thread applications, as race condition
+ * can be avoided.
+ *
+ * Available since: 0.2.4
+ */
+Bool
+trie_store_if_absent (Trie *trie, const AlphaChar *key, TrieData data)
+{
+    return trie_store_conditionally (trie, key, data, FALSE);
+}
+
+static Bool
+trie_store_conditionally (Trie            *trie,
+                          const AlphaChar *key,
+                          TrieData         data,
+                          Bool             is_overwrite)
 {
     TrieIndex        s, t;
     short            suffix_idx;
@@ -369,7 +444,10 @@ trie_store (Trie *trie, const AlphaChar *key, TrieData data)
             break;
     }
 
-    /* duplicated key, overwrite val */
+    /* duplicated key, overwrite val if flagged */
+    if (!is_overwrite) {
+        return FALSE;
+    }
     tail_set_data (trie->tail, t, data);
     trie->is_dirty = TRUE;
     return TRUE;
